"""Utilities."""
from base import Maybe, Unspecified, adict, as_mapping,\
        assignable_class_property, assignable_property,\
        autodocument_from_superclasses, class_cloner,\
        keyslice_pairs, lazy_property, luhn10,\
        named_int_factory, re_ucompile, re_uescape, symbol, threading, to_pairs

'''
from .deferred import deferred_module


deferred_module.shadow(
    'flatland.util',
    {
        'base': (
            'Maybe',
            'Unspecified',
            'adict',
            'as_mapping',
            'assignable_class_property',
            'assignable_property',
            'autodocument_from_superclasses',
            'class_cloner',
            'decorator',
            'format_argspec_plus',
            'keyslice_pairs',
            'lazy_property',
            'luhn10',
            'named_int_factory',
            're_ucompile',
            're_uescape',
            'symbol',
            'threading',
            'to_pairs',
            ),
        'deferred': (
            'deferred_module',
            ),
      })
'''
